// Test program for KMP
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

void KMP(unsigned char A[], unsigned long int n, unsigned char B[], unsigned long int m) {
  unsigned char * prefix = (unsigned char *) malloc(m * sizeof(char));
  unsigned long int curr;
  prefix[0] = 0;
  for (unsigned long int ind = 1; ind < m; ind++) {
    curr = prefix[ind-1];
    for (;;) {
      if (B[curr] == B[ind]) {
        prefix[ind] = curr+1;
        break;
      }
      if (curr == 0) {
        prefix[ind] = 0;
        break;
      }
      curr = prefix[curr-1];
    }
  }
  curr = 0;
  for (unsigned long int ind = 0; ind < n; ind++) {
    for (;;) {
      if (B[curr] == A[ind]) {
        curr++;
        break;
      }
      if (curr == 0) break;
      curr = prefix[curr-1];
    }
    if (curr == m) {
      // we found a match
      curr = prefix[curr-1];
    }
  }
  free(prefix);
}

void InitArrayWithRandomLetters(unsigned char A[], unsigned long int n) {
  for (unsigned long int k = 0; k < n; k++)
    A[k] = 'a' + rand()%2; // string consisting only of the letters 'a' and 'b' for more interesting text
}

unsigned long int GetCurrentTime() { // in clock ticks
  return clock();
}

void generateRandomModel (unsigned char A[], unsigned long int n, unsigned char B[], unsigned long int m, unsigned long int pos, unsigned long int matchingLength) {
  for (unsigned long int ind = 0; ind < matchingLength; ind++) {
    B[ind] = A[pos+ind];
  }
  for (unsigned long int ind = matchingLength; ind < m; ind++) {
    B[ind] = 'c'; // the text has only 'a' and 'b' so the pattern matching will end in 'c'
  }
}

unsigned long int generateRandomModels (unsigned char A[], unsigned long int n, unsigned char B[], unsigned long int m, unsigned long int matchingLength) {
  unsigned long int maxTime = 0;
  for (unsigned long int num = 0; num < 10; num++) {
    unsigned long int pos = (n-m)/9*num;
    if (num == 9) pos -= rand()%((n-m)/50);
    else if (num == 0) pos += rand()%((n-m)/50);
    else pos += rand()%((n-m)/50) - (n-m)/100;
    generateRandomModel(A,n,B,m,pos,matchingLength);
    unsigned long int before = GetCurrentTime();
    for (int r = 0; r < 5000; r++) KMP(A,n,B,m);
    unsigned long int after = GetCurrentTime();
    unsigned long int TimeElapsed = (after - before)/(CLOCKS_PER_SEC/100); // in centiseconds
    if (maxTime < TimeElapsed) maxTime=TimeElapsed;
  }
  return maxTime;
}

void TestAlgorithm() {
  unsigned char * A = (unsigned char *) malloc(30000L * sizeof(char));
  unsigned char * B = (unsigned char *) malloc(15000L * sizeof(char));
  for (int cnt = 1; cnt <= 10; cnt++) { // cnt = number of experiments
    for (unsigned int k = 1; k <= 10; k++) {
      if (k > 1) printf("\t");
      unsigned long int n = 3000L * k; // n = actual array length
      InitArrayWithRandomLetters(A,n);
      unsigned long int maxTime = 0;
      for (unsigned int s = 1; s <= 5; s++) {
        unsigned long int m = (n/10)*s;
        unsigned long int TimeElapsed;

        TimeElapsed = generateRandomModels(A,n,B,m,m);
        if (maxTime < TimeElapsed) maxTime=TimeElapsed;

        TimeElapsed = generateRandomModels(A,n,B,m,m-1);
        if (maxTime < TimeElapsed) maxTime=TimeElapsed;

        TimeElapsed = generateRandomModels(A,n,B,m,3*m/4);
        if (maxTime < TimeElapsed) maxTime=TimeElapsed;

        TimeElapsed = generateRandomModels(A,n,B,m,m/2);
        if (maxTime < TimeElapsed) maxTime=TimeElapsed;

        TimeElapsed = generateRandomModels(A,n,B,m,m/4);
        if (maxTime < TimeElapsed) maxTime=TimeElapsed;

        for (unsigned int r = 1; r <= 10; r++) {
          for (unsigned int ind = 0; ind < m; ind++) {
            B[ind] = 'a' + rand()%2;
          }
          unsigned long int before = GetCurrentTime();
          for (int r = 0; r < 5000; r++) KMP(A,n,B,m);
          unsigned long int after = GetCurrentTime();
          unsigned long int TimeElapsed = (after - before)/(CLOCKS_PER_SEC/100); // in centiseconds
          if (maxTime < TimeElapsed) maxTime=TimeElapsed;
        }
      }
      printf("%lu", maxTime);
    }
    printf("\n");
  }
  free(A);
  free(B);
}

int main() {
  TestAlgorithm();
}
